@@ -70,6 +70,8 @@ typedef struct json_create {
     unsigned int no_javascript_safe : 1;
     /* Make errors fatal. */
     unsigned int fatal_errors : 1;
+    /* Replace bad UTF-8 with the "replacement character". */
+    unsigned int replace_bad_utf8 : 1;
 }
 json_create_t;
 
@@ -305,129 +307,209 @@ add_u (json_create_t * jc, unsigned int u)
     }
 }
 
+#define BADUTF8								\
+    if (jc->replace_bad_utf8) {						\
+	/* We have to switch on Unicode otherwise the replacement */	\
+	/* characters don't work as intended. */			\
+	jc->unicode = 1;						\
+	CALL (add_str_len (jc, "\xEF\xBF\xBD", 3));			\
+    }									\
+    else {								\
+	json_create_user_message (jc, json_create_unicode_bad_utf8,	\
+				  "Invalid UTF-8");			\
+	return json_create_unicode_bad_utf8;				\
+    }
+
+/* Jump table. Doing it this way is not the fastest possible way, but
+   it's also very difficult for a compiler to mess this
+   up. Theoretically, it would be faster to make a jump table by the
+   compiler from the switch statement, but some compilers sometimes
+   cannot do that. */
+
+/* In this enum, I use three letters as a compromise between
+   readability and formatting. The control character names are from
+   "man ascii" with an X tagged on the end. */
+
+typedef enum {
+    CTL,  // control char, escape to \u
+    BSX,  // backslash b
+    HTX,  // Tab character
+    NLX,  // backslash n, new line
+    NPX,  // backslash f
+    CRX,  // backslash r
+    ASC,  // Non-special ASCII
+    QUO,  // double quote
+    BSL,  // backslash
+    FSL,  // forward slash, "/" 
+    BAD,  // Invalid UTF-8 value.
+    UT2,  // UTF-8, two bytes
+    UT3,  // UTF-8, three bytes
+    UT4,  // UTF-8, four bytes
+}
+jump_t;
+
+static jump_t jump[0x100] = {
+    CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,BSX,HTX,NLX,CTL,NPX,CRX,CTL,CTL,
+    CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,CTL,
+    ASC,ASC,QUO,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,FSL,
+    ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,
+    ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,
+    ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,BSL,ASC,ASC,ASC,
+    ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,
+    ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,ASC,
+    BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
+    BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
+    BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
+    BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
+    BAD,BAD,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,
+    UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,UT2,
+    UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,UT3,
+    UT4,UT4,UT4,UT4,UT4,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
+};
+
 /* Add a string to the buffer with quotes around it and escapes for
-   the escapables. When Unicode verification is added to the module,
-   it will be added here. */
+   the escapables. */
 
 static INLINE json_create_status_t
 json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN keylen)
 {
     int i;
+
     CALL (add_char (jc, '"'));
     for (i = 0; i < keylen; ) {
-	unsigned char c;
-	int j = i;
+	unsigned char c, d, e;
 	c = key[i];
-	if (c < 0x20) {
-	    if (c == '\t') {
-		ADD ("\\t");
-	    }
-	    else if (c == '\n') {
-		ADD ("\\n");
-	    }
-	    else if (c == '\r') {
-		ADD ("\\r");
-	    }
-	    else if (c == '\b') {
-		ADD ("\\b");
-	    }
-	    else if (c == '\f') {
-		ADD ("\\f");
-	    }
-	    else {
-		/* We know c is less than 0x10000, so we can use
-		   "add_one_u" not "add_u" here. */
-		CALL (add_one_u (jc, (unsigned int) c));
-	    }
+
+	switch (jump[c]) {
+
+	case CTL:
+	    CALL (add_one_u (jc, (unsigned int) c));
 	    i++;
-	}
-	else if (c < 0x80) {
-	    if (c == '"') {
-		ADD ("\\\"");
-	    }
-	    else if (c == '\\') {
-		ADD ("\\\\");
-	    }
-	    else if (c == '/' && jc->escape_slash) {
+	    break;
+
+	case BSX:
+	    ADD ("\\b");
+	    i++;
+	    break;
+
+	case HTX:
+	    ADD ("\\t");
+	    i++;
+	    break;
+
+	case NLX:
+	    ADD ("\\n");
+	    i++;
+	    break;
+
+	case NPX:
+	    ADD ("\\f");
+	    i++;
+	    break;
+
+	case CRX:
+	    ADD ("\\r");
+	    i++;
+	    break;
+
+	case ASC:
+	    CALL (add_char (jc, c));
+	    i++;
+	    break;
+
+	case QUO:
+	    ADD ("\\\"");
+	    i++;
+	    break;
+
+	case FSL:
+	    if (jc->escape_slash) {
 		ADD ("\\/");
 	    }
 	    else {
 		CALL (add_char (jc, c));
 	    }
 	    i++;
-	}
-	else {
-	    const unsigned char * input;
-	    input = key + i;
-	    if (c >= 0xf0 && c <= 0xf4) {
+	    break;
+
+	case BSL:
+	    ADD ("\\\\");
+	    i++;
+	    break;
+
+	case BAD:
+	    BADUTF8;
+	    i++;
+	    break;
+
+	case UT2:
+	    d = key[i + 1];
+	    if (d < 0x80 || d > 0xBF) {
+		BADUTF8;
+		i++;
+		break;
+	    }
+	    if (jc->unicode_escape_all) {
+		unsigned int u;
+		u = (c & 0x1F)<<6
+		  | (d & 0x3F);
+		CALL (add_u (jc, u));
+	    }
+	    else {
+		CALL (add_str_len (jc, (const char *) key + i, 2));
+	    }
+	    // Increment i
+	    i += 2;
+	    break;
+
+	case UT3:
+	    d = key[i + 1];
+	    e = key[i + 2];
+	    if (d < 0x80 || d > 0xBF ||
+		e < 0x80 || e > 0xBF) {
+		BADUTF8;
+		i++;
+		break;
+	    }
+	    if (! jc->no_javascript_safe &&
+		c == 0xe2 && d == 0x80 && 
+		(e == 0xa8 || e == 0xa9)) {
+		CALL (add_one_u (jc, 0x2028 + e - 0xa8));
+	    }
+	    else {
 		if (jc->unicode_escape_all) {
 		    unsigned int u;
-		    u = (input[0] & 0x07) << 18
-		      | (input[1] & 0x3F) << 12
-		      | (input[2] & 0x3F) <<  6
-		      | (input[3] & 0x3F);
-		    add_u (jc, u);
-		}
-		else {
-		    CALL (add_str_len (jc, (const char *) input, 2));
-		}
-		i += 4;
-	    }
-	    else if ((c & 0xE0) == 0xE0) {
-		/* Three byte case. */
-		if (input[1] < 0x80 || input[1] > 0xBF ||
-		    input[2] < 0x80 || input[2] > 0xBF) {
-		    json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
-		    return json_create_unicode_bad_utf8;
-		}
-		if (! jc->no_javascript_safe &&
-		    c == 0xe2 && input[1] == 0x80 && 
-		    (input[2] == 0xa8 || input[2] == 0xa9)) {
-		    CALL (add_one_u (jc, 0x2028 + input[2] - 0xa8));
-		    i += 3;
-		}
-		else {
-		    if (jc->unicode_escape_all) {
-			unsigned int u;
-			u = (input[0] & 0x0F)<<12
-			    | (input[1] & 0x3F)<<6
-			    | (input[2] & 0x3F);
-			CALL (add_u (jc, u));
-		    }
-		    else {
-			CALL (add_str_len (jc, (const char *) input, 3));
-		    }
-		}
-		i += 3;
-	    }
-	    else if ((c & 0xC0) == 0xC0) {
-		/* Two byte case. */
-		if (input[1] < 0x80 || input[1] > 0xBF) {
-		    json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
-		    return json_create_unicode_bad_utf8;
-		}
-		if (jc->unicode_escape_all) {
-		    unsigned int u;
-		    u = (input[0] & 0x1F)<<6
-		      | (input[1] & 0x3F);
+		    u = (c & 0x0F)<<12
+		      | (d & 0x3F)<<6
+		      | (e & 0x3F);
 		    CALL (add_u (jc, u));
 		}
 		else {
-		    CALL (add_str_len (jc, (const char *) input, 2));
+		    CALL (add_str_len (jc, (const char *) key + i, 3));
 		}
-		i += 2;
+	    }
+	    // Increment i
+	    i += 3;
+	    break;
+
+	case UT4:
+	    if (jc->unicode_escape_all) {
+		unsigned int u;
+		const unsigned char * input;
+		input = key + i;
+		u = (input[0] & 0x07) << 18
+		  | (input[1] & 0x3F) << 12
+		  | (input[2] & 0x3F) <<  6
+		  | (input[3] & 0x3F);
+		add_u (jc, u);
 	    }
 	    else {
-		json_create_user_message (jc, json_create_unicode_bad_utf8, "Invalid UTF-8");
-		return json_create_unicode_bad_utf8;
+		CALL (add_str_len (jc, (const char *) key + i, 4));
 	    }
-	}
-	if (j == i) {
-	    fprintf (stderr, "Failed to increment %d %d\n", i, j);
+	    // Increment i
+	    i += 4;
 	    break;
 	}
-	/* Unicode verification switch statements copied from
-	   JSON::Parse will go here. */
     }
     CALL (add_char (jc, '"'));
     return json_create_ok;
@@ -780,7 +862,6 @@ json_create_validate_user_json (json_create_t * jc, SV * json)
 static json_create_status_t
 json_create_call_to_json (json_create_t * jc, SV * cv, SV * r)
 {
-    I32 count;
     SV * json;
     char * jsonc;
     STRLEN jsonl;
@@ -1092,14 +1173,13 @@ json_create_run (json_create_t * jc, SV * input)
 /* Master routine, callers should only ever use this. Everything above
    is only for the sake of "json_create" to use. */
 
-static SV *
+static INLINE SV *
 json_create (SV * input)
 {
     /* With all the options, this really needs to be blanked out. Thus
        "buffer" is moved from being inside "jc" to being inside
        "json_create_run" above. */
     json_create_t jc = {0};
-
     /* Set up the default options. */
 
     /* Floating point number format. */
@@ -1110,10 +1190,16 @@ json_create (SV * input)
     jc.unicode_escape_all = 0;
     jc.handlers = 0;
     jc.type_handler = 0;
-
     return json_create_run (& jc, input);
 }
 
+/*  __  __      _   _               _     
+   |  \/  | ___| |_| |__   ___   __| |___ 
+   | |\/| |/ _ \ __| '_ \ / _ \ / _` / __|
+   | |  | |  __/ |_| | | | (_) | (_| \__ \
+   |_|  |_|\___|\__|_| |_|\___/ \__,_|___/ */
+                                       
+
 static json_create_status_t
 json_create_new (json_create_t ** jc_ptr)
 {
