@@ -7,7 +7,7 @@ require Exporter;
 );
 use warnings;
 use strict;
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 require XSLoader;
 XSLoader::load ('JSON::Create', $VERSION);
 
@@ -55,6 +55,15 @@ sub obj
     }
 }
 
+sub del
+{
+    my ($obj, @list) = @_;
+    my $handlers = $obj->get_handlers ();
+    for my $item (@list) {
+	delete $handlers->{$item};
+    }
+}
+
 sub validate
 {
     my ($obj, $value) = @_;
